# -*- coding: utf-8 -*-
"""
Created on Tue Sep 15 17:46:58 2020

@author: wuhaoyu
"""
import os
from datetime import datetime


def get_file_list(file_path):
    """
    按照创建日期对file_path下的文件进行排序

    Parameters
    ----------
    file_path : TYPE
        DESCRIPTION.

    Returns
    -------
    dir_list : TYPE
        DESCRIPTION.

    """
    dir_list = os.listdir(file_path)
    if not dir_list:
        return
    else:
        # 注意，这里使用lambda表达式，将文件按照最后修改时间顺序升序排列
        # os.path.getmtime() 函数是获取文件最后修改时间
        # os.path.getctime() 函数是获取文件最后创建时间
        dir_list = sorted(dir_list,key=lambda x: os.path.getctime(os.path.join(file_path, x)))
        return dir_list
    
    
def file_clean(file_folder, n=2):
    """
    清除file_folder目录下n年前历史文件

    Parameters
    ----------
    n : TYPE, optional
        DESCRIPTION. The default is 50.
    """
    if get_file_list(file_folder):
        file_list = [os.path.join(file_folder, i) for i in get_file_list(file_folder)]
    
        clean_list = filter(lambda x: datetime.fromtimestamp(os.path.getctime(x)).year == (datetime.now().year - n), file_list)
        for i in clean_list:
            os.remove(i)



import pandas as pd
import json
import os
from happyreport import excel_format
from happyreport import send_mail

class Setting(object):
    """
    替代settings.py的配置类
    """
    def __init__(self, job_url, jobs_from, data_url, save_path, host, port, **kwargs):
        """
        初始化属性
        """
        self.job_url = job_url
        self.jobs_from = jobs_from
        self.data_url = data_url
        self.save_path = save_path
        self.host = host
        self.port = port
        for key, value in kwargs.items():
            setattr(self, key, value)


class Setting2(object):
    """
    替代settings.py的配置类
    """
    def __init__(self, **kwargs):
        """
        初始化属性
        """
        for key, value in kwargs.items():
            setattr(self, key, value)


class MailJob(object):
    """
    报表获取，报表发送类
    执行逻辑：
        1、数仓构建源主题表数据：包含ds字段控制日期版本
        2、配置报表任务属性至数据库报表任务配置表：表包含字段：['job_id', 'job_name', 'table_fields', 'sender_email', 'password', 'receiver_email', 'acc_email', 'content']
            各字段释义：
                job_id              - int       : 报表任务ID
                job_name            - string    : 任务中文名即邮件主题名
                table_fields        - string    : JSON格式字符串，报表详细信息格式如下：
                        eg: [ {"workbook": [
                                    {"sheet": {
                                        "from": "主题表表名1",
                                        "select": ["字段1" , "字段2"],
                                        "as": ["别名1", "别名2"]},
                                     "sheet_name": "sheet页名称1"},
                                    {"sheet": {
                                        "from": "主题表表名2",
                                        "select": ["字段1" , "字段2"],
                                        "as": ["别名1", "别名2"]},
                                     "sheet_name": "sheet页名称2"}],
                               "attach_name": "附件名称.xlsx"},
                               {"workbook": [
                                    {"sheet": {
                                        "from": "主题表表名1",
                                        "select": ["字段1" , "字段2"],
                                        "as": ["别名1", "别名2"]},
                                     "sheet_name": "sheet页名称1"},
                                    {"sheet": {
                                        "from": "主题表表名2",
                                        "select": ["字段1" , "字段2"],
                                        "as": ["别名1", "别名2"]},
                                     "sheet_name": "sheet页名称2"}],
                               "attach_name": "附件名称.xlsx"}
                               ]
                sender_email        - string    : 发件人邮箱账号
                password            - string    : 发件人邮箱密码
                receiver_email      - string    : 收件人邮箱，多个收件人用“,“分隔
                acc_email           - string    : 抄送人邮箱，多个抄送人用“,“分隔
                content             - string    : 正文内容
        3、完成配置表实例化报表数据获取与邮箱发送类MailJob,结束任务
    """
    def __init__(self, job_id, settings, date):
        """
        初始化传入job_id与环境配置settings以及ds字段
        """
        self.date = date
        self.settings = settings
        self.jobs_from = settings.jobs_from
        self.jobs_url = settings.job_url
        self.data_url = settings.data_url
        self.job_id = job_id
        self.job = pd.read_sql(f"select * from {self.jobs_from} where job_id = '{job_id}'",
                               settings.job_url).to_dict(orient='series')
        self.job = dict(zip(self.job.keys(), map(lambda x: x.values[0], self.job.values())))
        self.attaches = json.loads(self.job['table_fields'])
        self.save_path = settings.save_path

    def _get_data(self):
        """
        获取报表数据，根据附件-子sheet，打包成list
        """
        res = []
        for at in self.attaches:
            sub = {"attach_name": at["attach_name"],
                   "data": [],
                   "sheet_names": []}
            for st in at['workbook']:
                sql = f"select {','.join(st['sheet']['select'])} from {st['sheet']['from']} where ds = '{self.date}'"
                print(sql)
                df = pd.read_sql(sql, self.data_url)
                df.columns = st['sheet']["as"]
                sub['data'].append(df)
                sub['sheet_names'].append(st['sheet_name'])
            res.append(sub)
        return res

    def _save(self, data_res):
        """
        保存附件至本地
        """
        for wb in data_res:
            excel_format(wb['data'], wb['sheet_names'], os.path.join(self.save_path, wb["attach_name"]))

    def _send_email(self):
        """
        根据任务配置项发送邮件
        """
        send_mail(host=self.settings.host,
                  port=self.settings.port,
                  user=self.job['sender_email'],
                  password=self.job['password'],
                  receivers=self.job['receiver_email'].split(","),
                  subject=self.job['job_name'],
                  carbon_copy=self.job['acc_email'].split(","),
                  carbon_copy_mask=self.job['acc_email'].split(","),
                  content=self.job['content'],
                  attaches=[os.path.join(self.save_path, a['attach_name']) for a in self.attaches])

    def run(self):
        """
        启动任务
        """
        data_res = self._get_data()
        self._save(data_res)
        self._send_email()



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 21 17:26:16 2020

@author: wuhaoyu

@description: 标准报表样式
"""

import pandas as pd
import numpy as np

def adaptive_column_width(data):
    """
    获取data最优列宽列表

    Parameters
    ----------
    data : TYPE
        待计算列宽的数据框
    Returns
    -------
    widths : TYPE
        对应列的最优列宽列表
    """
    widths = list(map(lambda c: max(np.median(list(map(len, map(str, data[c])))), len(c)) + 5, data))  
    return widths

def excel_format(data_tuple, filename, sheetname_tuple=None):
    """
    将传入的DataFrame元组按照模板的样式保存到指定路径filename

    Parameters
    ----------
    data_tuple : DataFrame的元组
        待保存的DataFrame.
    filename ：文件保存路径
        最终保存的工作簿
    sheetname_tuple : STR的元组
        最终保存的Sheet名.

    Returns
    -------
    None.

    """
    if not sheetname_tuple:
        sheetname_tuple = tuple(f"sheet{i+1}" for i in range(len(data_tuple)))
    with pd.ExcelWriter(filename) as writer:
        workbook = writer.book
        
        # 通用格式
        fmt = workbook.add_format({'font_name': '微软雅黑',
                                    'font_size': 10,
                                    'valign': 'vcenter',
                                    'align': 'center',
                                    })
        
        # 标题格式
        hfmt = workbook.add_format({'font_name': '微软雅黑',
                                    'font_size': 10,
                                    'align': 'vcenter',
                                    'valign': 'center',
                                    'bg_color': '#AED6F1',
                                    'bold': True,
                                    'text_wrap': True,
                                    'border': 1
                                    })
        
        # 数据写入工作表
        for data, sheetname in zip(data_tuple, sheetname_tuple):
            if len(data):
                datetime_columns = data.select_dtypes(['datetime']).columns
                data[datetime_columns] = data[datetime_columns].applymap(lambda x : str(x))
                data.to_excel(writer, sheet_name=sheetname, index=False)
                worksheet = writer.sheets[sheetname]
                
                # 自动调整列宽
                widths = adaptive_column_width(data)
                for i in range(len(widths)):
                    worksheet.set_column(f'{chr(65+i)}:{chr(65+i)}', widths[i], fmt)

                # 修改标题格式
                for col_num, value in enumerate(data.columns.values):
                    worksheet.write(0, col_num, value, hfmt)


from happyreport import send_mail, excel_format, file_clean
import custom
import settings

if __name__ == "__main__":
    # 获取报表保存
    wb = custom.workbook()

    # 格式化本地文件
    excel_format(data_tuple=wb, filename=settings.ATTACHES[0])

    # 发送邮件
    try:
        send_mail(host=settings.HOST,
                  port=settings.PORT,
                  user=settings.USER,
                  password=settings.PASSWORD,
                  receivers=settings.RECEIVERS,
                  subject=settings.SUBJECT,
                  attaches=settings.ATTACHES, content=settings.CONTENT)
    except Exception as e:
        send_mail(host=settings.HOST,
                  port=settings.PORT,
                  user=settings.USER,
                  password=settings.PASSWORD,
                  receivers=settings.WARNING_USER,
                  subject="【警告】邮件发送失败",
                  content=f'"{settings.SUBJECT}"发送失败, 报错信息:\n{str(e)}')

    # 清除历史文件
    file_clean("result")



# 邮箱必要参数
HOST = ''
PORT = 994
USER = ''
PASSWORD = ''
RECEIVERS = []
WARNING_USER = []
SUBJECT = ''
CONTENT = ''

# 邮箱可选参数
# 附件名
ATTACHES = []
SHEET_NAMES = []
# 抄件人
# CARBON_COPY = []
# MASK
# SENDER_MASK = ''
# RECEIVERS_MASK = []
# CARBON_COPY_MASK = []


# -*- coding: utf-8 -*-
"""
@author: wuhaoyu

@description：邮件发送模块
"""
import smtplib
import os
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.header import Header

def send_mail(host, port, user, password, receivers, subject, **kwargs):
    """
    host: 邮箱IP
    port: 邮箱服务端口
    user: 邮箱认证账号
    password: 认证账号密码
    receivers: 收件人列表
    subject: 邮件主题
    kwargs: from_mask\receivers_mask\carbon_copy\carbon_copy_mask\content\attaches
    """
    # 加载配置
    from_ = kwargs.get("from_mask", user)
    to = kwargs.get("receivers_mask", receivers)
    carbon_copy = kwargs.get("carbon_copy", [])
    carbon_copy_mask = kwargs.get("carbon_copy_mask", [])
    
    # 创建一个带邮件实例
    message = MIMEMultipart()
    message['From'] = Header(from_, 'utf-8')
    message['To'] = Header(",".join(to), 'utf-8')
    message['Cc'] = Header(",".join(carbon_copy_mask), 'utf-8')
    message['Subject'] = Header(subject, 'utf-8')

    # 邮件正文内容
    content = kwargs.get('content', '')
    message.attach(MIMEText(content, 'plain', 'utf-8'))
    
    # 附件子模块
    attaches = kwargs.get("attaches", None)
    if attaches:
        for attach in attaches:
            with open(attach, 'rb') as f:
                file = f.read()
            att = MIMEText(file, 'base64', 'utf-8')
            att["Content-Type"] = 'application/octet-stream'
            att.add_header("Content-Disposition", "attachment", filename=("gbk", "", os.path.split(attach)[-1]))
            message.attach(att)
    
    # 连接邮箱服务器
    smtp_obj = smtplib.SMTP_SSL(host, port)
    smtp_obj.login(user, password)
    try:
        smtp_obj.sendmail(user, list(set(receivers + carbon_copy)), message.as_string())
        print(f"邮件【{subject}】发送成功")
    except smtplib.SMTPException:
        print("Error: 无法发送邮件")
    finally:
        smtp_obj.close()



import os
import pandas as pd
import json
from json.decoder import JSONDecodeError

def check(settings):
    """
    环境与任务配置验证函数
    """
    columns = ['job_id', 'job_name', 'table_fields', 'sender_email', 'password', 'receiver_email', 'acc_email', 'content']
    df = pd.read_sql(settings.jobs_from, settings.job_url)
    missing_columns = []
    for c in df:
        if c not in columns:
            missing_columns.append(c)
    assert not missing_columns, f"配置表缺失标准字段{missing_columns}"

    for i in range(df.shape[0]):
        try:
            for wb in json.loads(df.iloc[i]['table_fields']):
                assert 'workbook' in wb.keys(), f"行号{i}table_fields缺少对象workbook"
                assert 'attach_name' in wb.keys(), f"行号{i}table_fields缺少对象attach_name"
                for st in wb['workbook']:
                    assert 'sheet' in wb.keys(), f"行号{i}workbook缺少对象sheet"
                    assert 'sheet_name' in st[k].keys(), f"行号{i}workbook缺少对象sheet_name"
                    for k, v in st.items():
                        assert 'from' in st[k].keys(), f"行号{i}sheet缺少对象from"
                        assert 'select' in st[k].keys(), f"行号{i}sheet缺少对象select"
                        assert 'as' in st[k].keys(), f"行号{i}sheet缺少对象as"
        except JSONDecodeError:
            raise JSONDecodeError(f"table_fields错误字段格式,行号：'{i}'")

    if not os.path(settings.save_path):
        print(f"创建报表保存路径{settings.save_path}")
        os.mkdir(os.path.join(os.path.split(settings.save_path)))
    print("验证通过")
